'''
CF-1.3 changes based on CF-1.2 checker code

implementation of requirements and recommendations from
http://cf-pcmdi.llnl.gov/conformance/requirements-and-recommendations/1.3/ and
http://cf-pcmdi.llnl.gov/documents/cf-conventions/1.3/cf-conventions.html

this code is an extension of the CF-1.0 checker base code in the cf10 module
and the modifications by the CF-1.1 and CF-1.2 checker codes in the cf11 and cf12
modules respectively
'''
import string

from utils import is_string_type, find_duplicates

from CFchecker import cf12

class CF13Tests(cf12.CF12Tests):
    '''
    changes compared to 1.2: new test 3.5 (see below for more information)
    '''
    cf_convention = 'CF-1.3'
    # there are no restrictions imposed on flag_meanings string contents before CF-1.5
    # but it seems reasonable that only printable characters should be included
    _allowed_flag_meanings_chars = set(string.printable)

    def _test_3_5(self):
        '''
        3.5 Flags
        * The flag_values attribute must have the same type as the variable to which it is attached.
        * The number of flag_values attribute values must equal the number of words or phrases
          appearing in the flag_meanings string.
        * The number of flag_masks attribute values must equal the number of words or phrases
          appearing in the flag_meanings string.
        * Variables with a flag_masks attribute must have a type that is compatible with bit
          field expression (char, byte, short and int), not floating-point (float, real, double),
          and the flag_masks attribute must have the same type.
        * The flag_masks attribute values must be non-zero.
        * The flag_values attribute values must be mutually exclusive among the set of
          flag_values attribute values defined for that variable.

        Recommendations:
        * When flag_masks and flag_values are both defined, the Boolean AND of each entry in
          flag_values with its corresponding entry in flag_masks should equal the flag_values entry,
          ie, the mask selects all the bits required to express the value.
        ADDITIONS:
        * only mentioned in table C1: standard_name attribute should have status_flag modifier
        * flag_meanings should only contain printable characters
        '''
        for vname, var in self.ncfile.variables.iteritems():
            std_name = getattr(var, 'standard_name', None)
            if not is_string_type(std_name): std_name = ''
            std_name_split = std_name.split()
            has_status_flag = len(std_name_split) == 2 and std_name_split[1] == 'status_flag'
            fvals_found = 'flag_values' in var.ncattrs()
            fmean_found = 'flag_meanings' in var.ncattrs()
            fmsk_found = 'flag_masks' in var.ncattrs()
            valid_types = ('S', 'b', 'h', 'i', 'l')
            # if flag_values or flag_meanings attributes are there without status_flag, warn about that
            if (fvals_found or fmean_found) and not has_status_flag:
                self.msg.warning('3.5', '%s: flag_* attributes found but status_flag modifier not set' % (vname,) )
            if fvals_found:
                fvals = getattr(var, 'flag_values')
                # string has no dtype and needs to be treated differently
                atttype = 'S' if is_string_type(fvals) else fvals.dtype.char
                vartype = self.ncfile.variables[vname].dtype.char
                # if flag_values is a string, we convert it to a list of ints for checks
                if atttype == 'S':
                    fvals = [ord(c) for c in fvals]
                fvals_len = len(fvals)
                # not strictly required for this attribute alone, but needed
                # if flag_masks is available too
                fvals_valid = atttype in valid_types
                if not vartype == atttype:
                    self.msg.error('3.5', '%s: flag_values attribute and variable need to have the same data type (%s <-> %s)' % (vname, atttype, vartype) )
                if not fmean_found:
                    self.msg.error('3.5', '%s: flag_values attribute without flag_meanings attribute' % (vname,) )
                dupe_fvals = find_duplicates(fvals)
                if dupe_fvals:
                    self.msg.error('3.5', '%s: every flag_values value has to be unique within the set of all flag_values' % (vname,) )
            if fmsk_found:
                fmsk = getattr(var, 'flag_masks')
                atttype = 'S' if is_string_type(fmsk) else fmsk.dtype.char
                vartype = self.ncfile.variables[vname].dtype.char
                # if flag_masks is a string, we convert it to a list of ints for checks
                if atttype == 'S':
                    fmsk = [ord(c) for c in fmsk]
                fmsk_len = len(fmsk)
                fmsk_valid = atttype in valid_types  # this will be needed later when ANDing
                if vartype not in valid_types:
                    self.msg.error('3.5', '%s: variable type must be compatible with bit field expressions (char, short, int, long)' % (vname,) )
                if not vartype == atttype:
                    self.msg.error('3.5', '%s: flag_masks attribute and variable need to have the same data type (%s <-> %s)' % (vname, atttype, vartype) )
                if 0 in fmsk:
                    self.msg.error('3.5', '%s: flag_masks attribute values must be non-zero' % (vname,) )
                if not fmean_found:
                    self.msg.error('3.5', '%s: flag_masks attribute without flag_meanings attribute' % (vname,) )
            if fmean_found:
                fmean = getattr(var, 'flag_meanings')
                if not is_string_type(fmean):
                    self.msg.error('3.5', '%s: flag_meanings attribute should be a string' % vname)
                else:
                    # flag_meanings is always a string at this point
                    fmean = fmean.split()
                    fmean_len = len(fmean)
                    # call outsourced check flag_meanings for invalid characters, it will write its own log entries
                    self._check_flag_meanings_phrases(vname, fmean)
                    if fvals_found:
                        if not fvals_len == fmean_len:
                            self.msg.error('3.5', '%s: flag_values and flag_meanings attributes have different lengths (%s <-> %s)' % (vname, fvals_len, fmean_len) )
                    if fmsk_found:
                        if not fmsk_len == fmean_len:
                            self.msg.error('3.5', '%s: flag_masks and flag_meanings attributes have different lengths (%s <-> %s)' % (vname, fmsk_len, fmean_len) )
                    if not fvals_found and not fmsk_found:
                        self.msg.error('3.5', '%s: flag_meanings attribute without flag_values or flag_masks attribute' % vname)
            if fvals_found and fmsk_found and fmsk_valid and fvals_valid:
                for val, msk in zip(fvals, fmsk):
                    if not val & msk == val:
                        self.msg.warning('3.5', '%s: boolean AND of flag_values and flag_masks entries should equal flag_values entry'
                                                % (vname,))
                        # one warning should be enough
                        break

    def _check_flag_meanings_phrases(self, vname, fmeanings):
        '''
        outsourced check for invalid characters in the flag_meanings attribute
        of test 3.5
        log messages will be written directly in this function instead of in test 3.5

        vname: name of variable that is being checked
        fmeanings: list of split flag_meanings phrases
        '''
        for phrase in fmeanings:
            invalid_chars = set(phrase) - self._allowed_flag_meanings_chars
            if invalid_chars:
                self.msg.error('3.5', '%s: flag_meanings phrase \'%s\' contains invalid characters: %s' \
                                      % (vname, phrase, ', '.join([("'%s'" % item) for item in invalid_chars])))
